/*
<trigger>
	<usage>bsf set_attribute.js -attr attribute_name -value value [-rel relationship_name] [-ifnotstate state_name]</usage>
	<version>5.2</version>
	<description>
		現在の CR、関係のある CR、または関連タスクの属性値を設定します。
	</description>
	<argument1>
		<name>-attr attribute_name</name>
		<required>必須</required>
		<type>string</type>
		<description>
			値を設定する対象の属性。
		</description>
	</argument1>
	<argument2>
		<name>-value value</name>
		<required>必須</required>
		<type>string</type>
		<description>
			設定する値。
		</description>
	</argument2>
	<argument3>
		<name>-rel relationship_name</name>
		<required>オプション</required>
		<type>string</type>
		<description>
			ターゲットの属性をもっているオブジェクトを指定します。
            指定しない場合は、現在の CR の属性に値が設定されます。
			"associated_task" とともに使用すると、すべての関連タスクの属性に値が設定されます。
			他の値の場合は、現在の CR と"relationship_name"関係をもつ CR の属性に値が設定されます。
		</description>
	</argument3>
	<argument4>
		<name>-if_not_state state_name</name>
		<required>no</required>
		<type>string</type>
		<description>
			ターゲット・オブジェクトが状態"state_name"を持つ場合には属性を設定しません。
			ターゲット・オブジェクトが CR の場合は、属性 crstatus をチェックします。
		    ターゲット・オブジェクトがタスクの場合は、属性 status をチェックします。
		</description>
	</argument4>
	<argument5>
		<name>-if_not_set true/false</name>
		<required>オプション</required>
		<type>boolean</type>
		<description>
			このパラメーターを 'true' とすると、ターゲット・オブジェクトがすでにこの属性に値を持っている場合は、属性を設定しません。
			値を一度設定してその後同じトリガーが二度以上実行されても値を変更しない場合に使います。
			このパラメーターを指定しない場合は、デフォルトは、'false' になります (属性はたとえ値をもっていても設定されます)。
		</description>
	</argument5>
</trigger>
*/


// Import packages
importPackage(Packages.com.telelogic.cs.api);

// Script specific global variables
const script_name = 'set_attribute.js';
const script_version = '20080620';
const script_usage = 'set_attribute.js -attr attribute_name -value value [-rel relationship_name] [-ifnotstate state_name]';
const script_example = 'set_attribute.js -attr estimate -value 0 -rel child_problem';
const log_header = 'Script bsf ' + script_name + ': ';
const script_debug = false;
const script_dcm_modifiable_check = true;

// Get objects
var args = bsf.lookupBean('args');
var event = bsf.lookupBean('event');
var change = bsf.lookupBean('api');
var log = bsf.lookupBean('log');
var options = OptionParser.parse(args);
var relationEvent = false;
if(event instanceof RelationEvent)
{
	relationEvent = true;
}

if (script_debug) 
{
	log.info(log_header);
	log.info('    Version: ' + script_version);
	log.info('    Arguments: ' + args.join(' '));
	log.info('    Current CR: ' + event.objectId);
}

if (checkArguments(0, ['-attr', '-value', '-rel', '-if_not_state', '-if_not_set'], ['-attr', '-value']))
{
	// Get Options
	var to = options.getOption('-attr');
	var value = options.getOption('-value');
	var if_not_state = options.getOption('-if_not_state');
	var rel = options.getOption('-rel');
	const ifNotSet = options.getOption('-if_not_set');
	var crid  = event.objectId;
	if (relationEvent)
	{
		crid = event.fromObjectId;
	}
	// Create attribute object for value
	var attribute = createAttribute(to, value);
	
	if (script_debug) 
	{
		if (isAttributeNull(attribute))
		{
			log.info(log_header + 'Value is null');
		}
		else
		{
			log.info(log_header + 'Value is \'' + attribute + '\' and is of type \'' + attribute.type + '\'');
		}
	}

	//
	// The script is from here to the end idential with copy_attribute.js
	//

	// Branch over relationship type
	var state_attribute_name;
	var attribute_names = [to];
	var items;
	
	if (!rel)
	{
		// Self
		if (script_debug)
		{
			log.info(log_header + 'Working on self');
		}

		state_attribute_name = 'crstatus';
	
		if (if_not_state) 
		{
			attribute_names.push[state_attribute_name];
		}

		if (script_dcm_modifiable_check)
		{
			attribute_names.push('modifiable_in');
		}
		
		items = change.query(event.triggerUser.token, change.buildCrQuery(event.triggerUser.token, crid), attribute_names);
	}
	else if (rel == 'associated_task')
	{
		// Associated tasks
		if (script_debug)
		{
			log.info(log_header + 'Working on associated tasks');
		}

		state_attribute_name = 'status';
		attribute_names.push('from_db_path');
		
		if (if_not_state) 
		{
			attribute_names.push(state_attribute_name);
		}
		
		if (script_dcm_modifiable_check)
		{
			attribute_names.push('modifiable_in');
		}
		
		items = change.getAssociatedTasks(event.triggerUser.token, crid, attribute_names);
	}
	else
	{
		// Associated CRs
		if (script_debug)
		{
			log.info(log_header + 'Working on associated CRs');
		}

		state_attribute_name = 'crstatus';
		
		if (if_not_state) 
		{
			attribute_names.push[state_attribute_name];
		}
		
		if (script_dcm_modifiable_check)
		{
			attribute_names.push('modifiable_in');
		}
		
		var has_string = 'has_'; 
		
		if (rel.indexOf(has_string) == 0)
		{
			items = change.getRelatedTo(event.triggerUser.token, event.toObjectId, rel.substr(has_string.length), attribute_names);
		}
		else
		{
			items = change.getRelatedFrom(event.triggerUser.token, event.fromObjectId, rel, attribute_names);
		}
	}
	
	// Filter list
	if (script_debug)
	{
		log.info(log_header + 'Item list: ' + items.join(', '));
	}

	items = filterItems(items, attribute, state_attribute_name, if_not_state, ifNotSet);
	
	if (script_debug) 
	{
		log.info(log_header + 'Filtered item list: ' + items.join(', '));
	}
	
	// Action item
	for (var i = 0; i < items.length; i++)
	{
		setObjectAttribute(items[i], attribute);
	}
}


/**
 * 
 * Checks the script arguments:
 * - Check argument count
 * - Check that all specified options are valid
 * - Check that all required options are specified
 * 
 * Returns: true, if succesfull. false otherwise
 */
function checkArguments(arg_count, allowed_options, required_options)
{
	var message;

	// Check argument count
	if (options.args.length < arg_count)
	{
		message = 'Insufficient arguments to run the script (' + options.args.length +' arguments)';
	}

	// Check that all specified options are valid
	if (!message)
	{
		var all_options = options.getOptions();
		var i = all_options.keySet().iterator();
	
		while (!message && i.hasNext())
		{
			var option = i.next();
			var is_valid = false;
		
			for (var j=0; j<allowed_options.length; j++)
			{
				if (option == allowed_options[j])
				{
					is_valid = true;
					break;
				}
			}
			
			if (!is_valid)
				message = 'Invalid option \'' + option + '\'';
		}
	}

	// Check that all required options are specified
	if (!message) 
	{
		for (var i = 0; i < required_options.length; i++)
		{
			var option = options.getOption(required_options[i]);
			
			if (!option)
			{
				message = 'Missing option \'' + required_options[i] + '\'';
			}
		}
	}

	if (message)
	{
		log.error(log_header);
		log.error('   ' + message);
		log.error('   Usage: ' + script_usage);
		log.error('   Example: ' + script_example);
		return false;
	}

	return true;
}

/**
 * Checks, if an attribute has no value set.
 * 
 * The function is necessary, because Attribute.toString() returns 'null' instead of null. By
 * returning 'null' it cannot be distinguished between an without a value and one with the 
 *  value 'null'.
 * 
 * Returns: True, if the attriute has no value set.
 */ 
function isAttributeNull(attribute)
{
	if (((!attribute) || (attribute.type == Attribute.Type.STRING) && (attribute.getStringValue() == null))
		|| ((attribute.type == Attribute.Type.DATE) && (attribute.getDateValue() == null))
		|| ((attribute.type == Attribute.Type.BOOLEAN) && (attribute.getBooleanValue() == null)))
	{
		return true;
	}
	
	return false;
}

/**
 * 
 * Creates a copy of a given attribute with a new name
 * 
 * The function is necessary, because new Attribute(name, null) will result in an error that
 * Attribute (java.lang.String,null) is ambiguous Attribute(java.lang.String,java.util.Date)
 * Attribute(java.lang.String,java.lang.Boolean) and Attribute(java.lang.String,java.lang.String).
 * 
 * Returns: New Attribute object
 */
function copyAttribute(name, attribute)
{
	if (isAttributeNull(attribute))
	{
		return new Attribute(name);
	}

	if (attribute.type == Attribute.Type.STRING)
	{
		return new Attribute(name, attribute.getStringValue());
	}
	
	if (attribute.type == Attribute.Type.DATE)
	{
		return new Attribute(name, attribute.getDateValue());
	}
	
	if (attribute.type == Attribute.Type.BOOLEAN)
	{
		return new Attribute(name, attribute.getBooleanValue());
	}

	log.error(log_header + 'Found unknown attribute type: ' + attribute.type);
	
	return null;
}

/**
 * 
 * Creates an attribute object for the attribute of the specified name
 * 
 * The function is necessary to ensure that the attribute object has the correct type.
 * If created with the wrong type the attribute is cleared.
 * 
 * Returns: New Attribute object
 */
function createAttribute(name, value)
{
	if (!name)
	{
		return null;
	}

	if (!value)
	{
		return new Attribute(name);
	}

	// Lookup attribute type in the lifecycle
	var attributedefs = change.getAttributeDefs(event.triggerUser.token);
	var type;
	
	for (var i = 0; i < attributedefs.length; i++)
	{
		if (attributedefs[i].name.equals(name)) // == is always false
		{
			type = attributedefs[i].type;
			break;
		}
	}

	if (!type)
	{
		// Do not write an error. It could be an attribute for a task
		if (script_debug) 
		{
			log.info(log_header + 'Attribute ' + name + ' has no attribute definition in the lifecycle. Using string instead');
		}

		return new Attribute(name, value.toString());
	}

	// The type will be the one from the lifecycle editor not Attribute.Type
	if (type == 'time')
	{
		value = new Date(value);
	}
	else if (type == 'boolean')
	{
		value = new Boolean(value);
	}
	else
	{
		value = value.toString();
	}

	return new Attribute(name, value);
}

/**
 * Filters items from the array of items to be handled by the script
 * - The to attribute has already the desired value
 * - The state is the state to be excluded
 * - The item is not modifiable in the current database and script_dcm_modifiable_check is true
 * 
 * Returns: Filtered list of items
 */
function filterItems(items, attribute, state_attribute_name, if_not_state, ifNotSet)
{
	var new_items = new Array();
	
	for (var i = 0; i < items.length; i++)
	{
		// The to attribute has already the desired value
		// Check explicitely for empty attributes (see isAttributeNull for details) 
		if ((isAttributeNull(items[i].getAttribute(attribute.getName())) && isAttributeNull(attribute)) || 
			(!isAttributeNull(items[i].getAttribute(attribute.getName())) && !isAttributeNull(attribute) &&
			 items[i].getAttribute(attribute.getName()).toString().equals(attribute.toString())) // == is always false
			)
		{
			if (script_debug) 
			{
				log.info(log_header + 'Skipping equal value \'' + attribute.toString() + '\' in item ' + items[i]);
			}
			
			continue;
		}

		if (ifNotSet)
		{
			// Don't reset an attribute is it's already set to some value.
			if (items[i].getAttribute(attribute.getName()))
			{
				if (script_debug)
				{
					log.info(log_header + 'Skipping already set value \'' + attribute.toString() + '\' in item ' + items[i]);
				}

				continue;
			}
		}

		// The state is the state to be excluded
		if (items[i].getAttribute(state_attribute_name) && 
			if_not_state && 
			items[i].getAttribute(state_attribute_name).toString().equals(if_not_state)) // == is always false
		{
			if (script_debug) 
			{
				log.info(log_header + 'Skipping exluded state \'' + if_not_state + '\' in item ' + items[i]);
			}
			
			continue;
		}

		// The item is not modifiable in the current database and script_dcm_modifiable_check is true
		if (script_dcm_modifiable_check &&
			change.isDcmEnabled(event.triggerUser.token) &&
			!items[i].getAttribute('modifiable_in').toString().equals(change.getDbid(event.triggerUser.token)) &&
			!(change.isCentralServer(event.triggerUser.token) && items[i] instanceof Task)) // == is always false
		{
			if (script_debug) 
			{
				log.info(log_header + 'Skipping non modifiable item. Item \'' + items [i]+ 
				'\' is not modifiable in database \'' + change.getDbid(event.triggerUser.token) + '\'');
			}
			
			continue;
		}

		new_items.push(items[i]);
	}

	return new_items;
}


/*
 * 
 * Sets an attribute on a CR or task
 *
 */
function setObjectAttribute(item, attribute)
{
	var is_task = (item.toString().indexOf('Task') == 0);
	var attributes = [attribute];
	var item_id_string;

	try 
	{
		if (is_task)
		{
			// Set attribute on a task
			item_id_string = 'Task ' +  item.taskId;
			if (script_debug) 
			{
				log.info(log_header + 'Modifying ' + item_id_string);
			}
	
			// Override the context database in central server mode
			var from_db_path = item.getAttribute('from_db_path');
			if (from_db_path)
			{
				attributes.push[from_db_path];
			}
	
			change.setTaskAttributes(event.triggerUser.token, item.taskId, attributes);	
		}
		else 
		{
			// Set attribute on a CR
			item_id_string = 'CR ' +  item.crid;
			if (script_debug) 
			{
				log.info(log_header + 'Modifying ' + item_id_string);
			}
	
			change.setAttributes(event.triggerUser.token, item.crid, attributes);	
		}
	}
	catch (error)
	{
		var message = change.getMessage('errorSettingAttributeXonItemYerrorZ', [attribute.name, item_id_string, error.message || error.toString()]);
		log.error(log_header + 'Copying from CR ' + event.objectId + '. ' + message); 
		event.fail(message);
	}
}
